#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test_clustering.py: Test the clustering module
"""

from vulpes.automl import Clustering

import numpy as np


def test_clustering_1() -> None:
    # clustering
    X = np.array(
        [
            [
                np.random.randint(-20, 20) * np.random.random(),
                np.random.randint(-10, 10) * np.random.random(),
            ]
            for _ in range(50)
        ]
    )
    clustering = Clustering()
    _ = clustering.fit(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Top-level package for Vulpes."""

__author__ = "Adrien Carrel"
__email__ = "a.carrel@hotmail.fr"
__version__ = "0.2.0"

__all__ = ["utils",
           "automl"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""utils.py: Core functions and objects of the Vulpes package.

@Author: Adrien Carrel
"""

import warnings
from typing import List, Any, Union

import numpy as np
import pandas as pd
import xgboost
import lightgbm
from sklearn.utils import all_estimators
from sklearn.metrics import (
    make_scorer,
    accuracy_score,
    balanced_accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    mean_absolute_error,
    r2_score,
    mean_squared_error,
    roc_auc_score,
    precision_recall_curve,
    auc,
    mean_absolute_percentage_error,
    average_precision_score,
    pairwise_distances,
    calinski_harabasz_score,
    davies_bouldin_score,
    silhouette_score,
)
from sklearn.preprocessing import label_binarize
from sklearn.utils.extmath import softmax

# define type Array_like
Array_like = Union[List, pd.DataFrame, pd.Series, np.ndarray, Any]


def pr_auc_score(y: np.ndarray, y_pred: np.ndarray, **kwargs) -> float:
    """
    Function to calculate the PR AUC Score for
    binary and multiclass classification.

    Args:
        y (np.ndarray): True labels.
        y_pred (np.ndarray): Target scores (e.g. probability).
        kwargs: optional keyword arguments of
            sklearn.metrics.precision_recall_curve

    Returns:
        float: Area under the ROC curve.

    Examples:
        >>> from vulpes import pr_auc_score
        >>> y_true = np.array([0, 0, 1, 1])
        >>> y_scores = np.array([0.1, 0.4, 0.35, 0.8])
        >>> pr_auc_score(y_true, y_scores)
    """
    if len(y_pred.shape) == 1:  # binary classification
        precision, recall, _ = precision_recall_curve(y, y_pred, **kwargs)
    else:
        classes = list(range(y_pred.shape[1]))
        if len(classes) == 2:  # binary classification too
            precision, recall, _ = precision_recall_curve(y, y_pred[:, 1], **kwargs)
        else:  # multiclass
            Y = label_binarize(y, classes=classes)
            precision, recall, _ = precision_recall_curve(
                Y.ravel(), y_pred.ravel(), **kwargs
            )
    return auc(recall, precision)


def avg_precision(y: np.ndarray, y_pred: np.ndarray, **kwargs) -> float:
    """
    Micro-average precision for binary and
    multiclass.

    Calculate metrics globally by considering each element of the label
    indicator matrix as a label.

    Args:
        y (np.ndarray): True labels.
        y_pred (np.ndarray): Target scores (e.g. probability).
        kwargs: optional keyword arguments of
            sklearn.metrics.average_precision_score

    Returns:
        float: Micro average precision score.

    Examples:
        >>> from vulpes import avg_precision
        >>> y_true = np.array([0, 0, 1, 1])
        >>> y_scores = np.array([0.1, 0.4, 0.35, 0.8])
        >>> avg_precision(y_true, y_scores)
    """
    if len(y_pred.shape) == 1:  # binary classification
        return average_precision_score(y, y_pred, average="micro", **kwargs)
    classes = list(range(y_pred.shape[1]))
    if len(classes) == 2:  # binary classification too
        return average_precision_score(y, y_pred[:, 1], average="micro", **kwargs)
    # multiclass
    Y = label_binarize(y, classes=classes)
    return average_precision_score(Y, y_pred, average="micro", **kwargs)


# metrics that will be calculated for each classification models
CUSTOM_SCORER_CLF = {
    "balanced_accuracy": make_scorer(balanced_accuracy_score, greater_is_better=True),
    "accuracy": make_scorer(accuracy_score, greater_is_better=True),
    "precision": make_scorer(precision_score, average="macro"),
    "recall": make_scorer(recall_score, average="macro"),
    "f1": make_scorer(f1_score, average="macro", greater_is_better=True),
    "auroc": make_scorer(
        roc_auc_score,
        multi_class="ovo",
        average="macro",
        needs_proba=True,
        greater_is_better=True,
    ),
    "auprc": make_scorer(pr_auc_score, needs_proba=True, greater_is_better=True),
    "avg_precision": make_scorer(
        avg_precision, needs_proba=True, greater_is_better=True
    ),
}


def r2_score_adj(
    y: np.ndarray, y_pred: np.ndarray, *, n: int, p: int, fit_intercept: bool = True
) -> float:
    """
    Calculate Adjusted R2 Score.

    Adapted from:
    https://stackoverflow.com/questions/69901671/
    how-to-create-an-adjusted-r-squared-scorer-using-sklearn-metrics-make-scorer

    Args:
        y (np.ndarray): True labels.
        y_pred (np.ndarray): Target scores (e.g. probability).
        n (int): Number of samples.
        p (int): Number of parameters.
        fit_intercept (bool, optional): Whether of not we fitted the intercept.
            Defaults to True.

    Returns:
        float: Adjusted R2 Score based on the true values and the predicted
            values.

    Examples:
        >>> from vulpes import r2_score_adj
        >>> y_true = np.array([0, 1, 2, 3, 4])
        >>> y_scores = np.array([-0.01, 1.05, 1.98, 3.12, 3.93])
        >>> r2_score_adj(y_true, y_scores, n=5, p=2, fit_intercept=True)
    """
    if fit_intercept:
        rsquared = 1 - np.nansum((y - y_pred) ** 2) / np.nansum(
            (y - np.nanmean(y)) ** 2
        )
        rsquared_adj = 1 - ((n - 1) / (n - p - 1)) * (1 - rsquared)
    else:
        rsquared = 1 - np.nansum((y - y_pred) ** 2) / np.nansum(y**2)
        rsquared_adj = 1 - (n / (n - p)) * (1 - rsquared)
    return rsquared_adj


# metrics that will be calculated for each regression models

# adj r2 scorer is None, will be modified in the
# pipeline below (as we need to retrieve the n and p parameters).
# It's an approximation that doesn't take into account the
# change of n when splitting the dataset
CUSTOM_SCORER_REG = {
    "r2": make_scorer(r2_score, greater_is_better=True),
    "rmse": make_scorer(mean_squared_error, squared=True, greater_is_better=False),
    "mae": make_scorer(mean_absolute_error, greater_is_better=False),
    "mape": make_scorer(mean_absolute_percentage_error, greater_is_better=False),
    "adj_r2": None,
}

# metrics that will be calculated for each clustering algorithms
# Davies–Bouldin Index (DBI), lower is better
CUSTOM_SCORER_CLT = {
    "calinski_harabasz": calinski_harabasz_score,
    "silhouette": silhouette_score,
    "davies_bouldin": davies_bouldin_score,
}

# Dictionnary with prettier names for the metrics (to print the final result)
METRIC_NAMES = {
    "balanced_accuracy": "Balanced Accuracy",
    "accuracy": "Accuracy",
    "recall": "Recall",
    "precision": "Precision",
    "f1": "F1 Score",
    "auroc": "AUROC",
    "auprc": "AUPRC",
    "avg_precision": "Micro avg Precision",
    "r2": "R2",
    "rmse": "RMSE",
    "mae": "MAE",
    "mape": "MAPE",
    "adj_r2": "Adjusted R2",
    "calinski_harabasz": "Calinski-Harabasz Index",
    "silhouette": "Mean Silhouette Coefficient",
    "davies_bouldin": "Davies–Bouldin Index",
}

# Metrics to reverse (because lower is better)
# ex: rmse, mae, mape
METRICS_TO_REVERSE = ["RMSE", "MAE", "MAPE"]

# Extract classifiers, regressions and clustering algorithms from scikit learn
CLASSIFIERS = [est for est in all_estimators(type_filter=["classifier"])]
REGRESSIONS = [est for est in all_estimators(type_filter=["regressor"])]
CLUSTERING = [est for est in all_estimators(type_filter=["cluster"])]
# Manually add some others
CLASSIFIERS.append(("XGBClassifier", xgboost.XGBClassifier))
CLASSIFIERS.append(("LGBMClassifier", lightgbm.LGBMClassifier))
REGRESSIONS.append(("XGBRegressor", xgboost.XGBRegressor))
REGRESSIONS.append(("LGBMRegressor", lightgbm.LGBMRegressor))
# Remove voting classifiers, multi-task, etc
CLASSIFIERS_TO_REMOVE = set(
    [
        "CategoricalNB",
        "ClassifierChain",
        "ComplementNB",
        "GaussianProcessClassifier",
        "MultiOutputClassifier",
        "MultinomialNB",
        "NuSVC",
        "OneVsOneClassifier",
        "OneVsRestClassifier",
        "OutputCodeClassifier",
        "PassiveAggressiveClassifier",
        "RadiusNeighborsClassifier",
        "StackingClassifier",
        "VotingClassifier",
    ]
)
CLASSIFIERS = [clf for clf in CLASSIFIERS if clf[0] not in CLASSIFIERS_TO_REMOVE]
# Remove voting regressions, multi-task, etc
REGRESSIONS_TO_REMOVE = set(
    [
        "CCA",
        "IsotonicRegression",
        "GammaRegressor",
        "MultiOutputRegressor",
        "MultiTaskElasticNet",
        "MultiTaskElasticNetCV",
        "MultiTaskLasso",
        "MultiTaskLassoCV",
        "PoissonRegressor",
        "PLSCanonical",
        "PLSRegression",
        "QuantileRegressor",
        "RadiusNeighborsRegressor",
        "RegressorChain",
        "StackingRegressor",
        "VotingRegressor",
    ]
)
REGRESSIONS = [reg for reg in REGRESSIONS if reg[0] not in REGRESSIONS_TO_REMOVE]
# Clustering algorithms to remove
CLUSTERING_TO_REMOVE = set(["FeatureAgglomeration"])
CLUSTERING = [clt for clt in CLUSTERING if clt[0] not in CLUSTERING_TO_REMOVE]


def sigmoid_(x: float) -> float:
    """
    Calculate the sigmoid of x

    Args:
        x (float): float

    Returns:
        float: sigmoid(x)

    Examples:

    """
    return 1 / (1 + np.exp(-x))


def sigmoid_array(x: Union[np.ndarray, List]) -> np.ndarray:
    """
    Based on a list of values, calculate two-class probabilities
    using the sigmoid function.

    Args:
        x (Union[np.ndarray, List]): value (e.g. result
        of the decision function)

    Returns:
        np.ndarray: An array with probabilities for a two class
        classification problem

    Examples:

    """
    return np.array(list(map(lambda e: [sigmoid_(-e), sigmoid_(e)], x)))


def create_model_2(model: Any) -> Any:
    """
    Create a parent class of the model "model"
    to add a .predict_proba method based on either the
    decision function, or pairwise distances

    Args:
        model (Any): instance of model object

    Returns:
        Any: extended class with a new method

    Examples:
        >>> from vulpes import create_model_2
        >>> from sklearn.neighbors import NearestCentroid
        >>> model = NearestCentroid
        >>> model = create_model_2(model)
        >>> model.fit([[1, 2], [2, 3], [3, 4], [4, 5]], [0, 0, 1, 1])
        >>> model.predict_proba([[2, 2]])
        array([[0.90099855, 0.09900145]])
    """
    if hasattr(model, "decision_function"):
        # if there is a decision function, use it to calculate proba
        class model2(model):
            def __init__(self):
                super().__init__()

            def predict_proba(self, X: Array_like):
                d = self.decision_function(X)
                if len(d.shape) == 1:  # probably two outputs
                    return sigmoid_array(d)
                return softmax(d)

    # no decision function, calculate pairwise distance to clusters instead
    # exemple: NearestCentroid. But need centroids
    else:

        class model2(model):
            def __init__(self):
                super().__init__()

            def predict_proba(self, X: Array_like):
                distances = pairwise_distances(
                    X, Y=self.centroids_, metric="euclidean", n_jobs=-1
                )
                return softmax(-distances)

    return model2()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test_others.py: Test the other functions
"""

from vulpes.automl import Classifiers

import numpy as np


def test_others_1() -> None:
    X = np.array(
        [
            [
                np.random.randint(-20, 20) * np.random.random(),
                np.random.randint(-10, 10) * np.random.random(),
            ]
            for _ in range(50)
        ]
    )
    X[0][1] = np.nan
    X[1][0] = None
    classifiers = Classifiers()
    _ = classifiers.missing_data(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""clustering.py: Class Clustering
to test many clustering algorithms
"""

from .corevulpes import CoreVulpes
from ..utils.utils import CUSTOM_SCORER_CLT, METRIC_NAMES, METRICS_TO_REVERSE

import warnings
from time import perf_counter
from typing import List, Dict, Any, Union, Tuple
from collections import defaultdict
from os.path import join as opj

import numpy as np
import pandas as pd
from tqdm import tqdm
from sklearn.pipeline import Pipeline

# define type Array_like
Array_like = Union[List, pd.DataFrame, pd.Series, np.ndarray, Any]


class Clustering(CoreVulpes):
    """
    Object to train many regressions.
    All the parameters are optionals and can be modified.

    Args:
        models_to_try (Union[str, List[Tuple[str, Any]]], optional):
            List of models to try. It can be either a string that corresponds
            to a predefined list of models ("all", ...) or it can be a list
            of tuple (name of a model, class of a model)
            (e.g. ("KMeans",
                   sklearn.cluster.KMeans)).
            Defaults to "all" (train all the available clustering
            algorithms).
        custom_scorer (Dict[str, Any], optional): metrics to calculate after
            fitting a model. Dictionary with pairs name:scorer where the
            scorer is created using the function make_scorer from sklearn.
            Defaults to CUSTOM_SCORER_CLT.
        preprocessing (Union[Pipeline, str], optional): preprocessing
            pipeline to use. It can be None (no preprocessing), a
            predefined preprocessing pipeline ("default", ...) or
            a Pipeline object from sklearn. Defaults to "default":
            it applies a OneHotEncoder to "category" and object features,
            and it applies a SimpleImputer (median strategy) and a
            StandardScaler to numerical features.
        sort_result_by (str, optional): on which metric do you want to
            sort the final dataframe. Defaults to "Davies–Bouldin Index".
        ascending (bool, optional): sort the final dataframe in
            ascending order?. Defaults to True.
        save_results (bool, optional): if True, save the results in a csv
            file. Defaults to False.
        path_results (str, optional): path to use when saving the results.
            Defaults to "".
        additional_model_params (Dict[str, Any], optional): dictionary
            that contains parameters to be applied to each element of the
            pipeline. E.g. {"n_estimators": 100}, apply to all the
            preprocessing tasks and/or models that have the parameter
            n_estimators with the parameter n_estimators.
            Defaults to {"nb_clusters": 3, "min_samples": 5, "eps": 0.5}.
        random_state (int, optional): random state variable. Is applied
            to every model and elements of the pipeline. Defaults to 42.
        verbose (int, optional): if greater than 1, print the warnings.
            Defaults to 0.

    Examples:

        >>> from sklearn.datasets import load_iris
        >>> from vulpes.automl import Clustering

        >>> dataset = load_iris()
        >>> X = pd.DataFrame(dataset["data"], columns=dataset["feature_names"])

        >>> clustering = Clustering()
        >>> df_models = clustering.fit(X)
        >>> df_models
        | Model                   | Calinski-Harabasz Index | ...
        |-------------------------|-------------------------|----------
        | AgglomerativeClustering | 502.821564		        | ...
        | MeanShift	              | 509.703427		        | ...
        | Birch                   | 458.472511	            | ...
        | SpectralClustering	  | 410.369441	            | ...
        | ...                     | ...                     | ...
    """
    def __init__(
        self,
        *,
        models_to_try: Union[str, List[Tuple[str, Any]]] = "all",
        custom_scorer: Dict[str, Any] = CUSTOM_SCORER_CLT,
        preprocessing: Union[Pipeline, str] = "default",
        sort_result_by: str = "Davies–Bouldin Index",
        ascending: bool = True,
        save_results: bool = False,
        path_results: str = "",
        additional_model_params: Dict[str, Any] = {"nb_clusters": 3,
                                                   "min_samples": 5,
                                                   "eps": 0.5},
        random_state: int = None,
        verbose: int = 0,
    ):
        super().__init__()
        self.task = "clustering"
        self.models_to_try = self.predefined_list_models(models_to_try)
        self.custom_scorer = custom_scorer
        self.preprocessing = preprocessing
        self.sort_result_by = sort_result_by
        self.ascending = ascending
        self.save_results = save_results
        self.path_results = path_results
        self.additional_model_params = additional_model_params
        self.random_state = random_state
        self.verbose = verbose

        if not (self.verbose):
            warnings.filterwarnings("ignore")

    def fit(
        self,
        X: Array_like,
        *,
        sample_weight: Array_like = None
    ) -> pd.DataFrame:
        """
        Fit many clustering algorithms

        Args:
            X (Array_like): Input dataset
            sample_weight (Array_like, optional): sample weights
                to apply when fitting. Defaults to None.

        Raises:
            RuntimeError: Error when fitting a model

        Returns:
            pd.DataFrame: dataframe with the goodness-of-fit metrics
                evaluated for each clustering algorithm.

        Examples:

        """
        # Convert X to dataframe
        # (some preprocessing task, model, etc require this format)
        if not (isinstance(X, pd.DataFrame)):
            X = pd.DataFrame(X)

        # dictionary to store calculated values, model info, etc for each model
        metrics_dic = defaultdict(list)

        # Fit and calculate metrics for all the models
        for name, model in tqdm(self.models_to_try):
            print("\n" + name)
            top = perf_counter()

            model = model()
            model_name = name.lower()
            # Create the pipeline
            pipe = self.create_pipeline(model_name, model)

            # When available:
            # - maximize n_jobs
            # - fix a random_state
            model_params = {}
            for pipe_name, pipe_elt in pipe.steps:
                pipe_elt_available_params = pipe_elt.get_params().keys()
                if "random_state" in pipe_elt_available_params:
                    rd_state = f"{pipe_name}__random_state"
                    model_params[rd_state] = self.random_state
                if "normalize" in pipe_elt_available_params:
                    model_params[f"{pipe_name}__normalize"] = False
                if "n_jobs" in pipe_elt_available_params:
                    model_params[f"{pipe_name}__n_jobs"] = -1
                for (
                    add_param_name,
                    add_param_val,
                ) in self.additional_model_params.items():
                    if add_param_name in pipe_elt_available_params:
                        model_params[
                            f"{pipe_name}__{add_param_name}"
                        ] = add_param_val
            if model_params != {}:
                pipe.set_params(**model_params)

            # if sample_weight is a parameter of the fit function, update
            # fit parameters
            fit_params = {}
            for pipe_name, pipe_elt in pipe.steps:
                if hasattr(pipe_elt, "fit") and (
                    "sample_weight" in pipe_elt.fit.__code__.co_varnames
                ):
                    fit_params[f"{pipe_name}__sample_weight"] = sample_weight

            try:
                pipe.fit(X, **fit_params)
            except Exception as e:
                raise RuntimeError(f"Error when fitting: {e}")
            res = {}
            for name_scorer, scorer in self.custom_scorer.items():
                try:
                    # pipe[-1] because (clustering) model at the end
                    score_val = scorer(X, pipe[-1].labels_)
                except Exception as e:
                    print(f"Error when calculating {name_scorer} on {name}")
                    print(e)
                    score_val = np.nan
                res[name_scorer] = score_val

            # store fitted models
            self.fitted_models_[name] = pipe

            # add metrics to the lists

            # nan mean are used, because when folding,
            # a label can be missing
            # in a particular fold, thus creating nan values
            for metric_name in self.custom_scorer.keys():
                # print_metric_name (pmn)
                pmn = METRIC_NAMES.get(metric_name, metric_name)
                (metrics_dic[pmn].append(np.nanmean(res[metric_name])))
            metrics_dic["Model"].append(name)  # add name
            # add running time
            metrics_dic["Running time"].append(perf_counter() - top)

        # reverse these metrics (because lower is better)
        # ex: rmse, mae, mape
        for metric_name in METRICS_TO_REVERSE:
            if metric_name in metrics_dic:
                reverse_metric = [-x for x in metrics_dic[metric_name]]
                metrics_dic[metric_name] = reverse_metric
        df_models = pd.DataFrame.from_dict(metrics_dic)
        df_models = df_models.sort_values(
            by=self.sort_result_by, ascending=self.ascending
        ).set_index("Model")
        self.df_models = df_models

        if self.save_results:  # save results
            df_models.to_csv(
                opj.path.join(self.path_results,
                              f"vulpes_results_{self.task}.csv")
            )
        return df_models



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .utils import REGRESSIONS_TO_REMOVE, Array_like
from .utils import CUSTOM_SCORER_CLF
from .utils import CUSTOM_SCORER_REG
from .utils import CUSTOM_SCORER_CLT
from .utils import METRIC_NAMES
from .utils import METRICS_TO_REVERSE
from .utils import CLASSIFIERS
from .utils import REGRESSIONS
from .utils import CLUSTERING
from .utils import pr_auc_score
from .utils import avg_precision
from .utils import r2_score_adj
from .utils import sigmoid_
from .utils import sigmoid_array
from .utils import create_model_2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""classifiers.py: Class Classifiers
to test many classification models
"""

from .corevulpes import CoreVulpes
from ..utils.utils import (
    CUSTOM_SCORER_CLF,
    METRIC_NAMES,
    METRICS_TO_REVERSE,
    create_model_2,
)

import warnings
import numbers
from time import perf_counter
from typing import List, Dict, Any, Union, Tuple
from collections import defaultdict
from collections.abc import Iterable
from os.path import join as opj

import numpy as np
import pandas as pd
from tqdm import tqdm
from sklearn.pipeline import Pipeline
from sklearn.model_selection import (
    cross_validate,
    RepeatedKFold,
    _validation,
    train_test_split,
)

# define type Array_like
Array_like = Union[List, pd.DataFrame, pd.Series, np.ndarray, Any]


class Classifiers(CoreVulpes):
    """
    Object to train many classifiers.
    All the parameters are optionals and can be modified.

    Args:
        models_to_try (Union[str, List[Tuple[str, Any]]], optional):
            List of models to try. It can be either a string that corresponds
            to a predefined list of models ("all", ...) or it can be a list
            of tuple (name of a model, class of a model)
            (e.g. ("RandomForestClassifier",
                   sklearn.ensemble.RandomForestClassifier)).
            Defaults to "all" (train all the available classification
            algorithms).
        custom_scorer (Dict[str, Any], optional): metrics to calculate after
            fitting a model. Dictionary with pairs name:scorer where the
            scorer is created using the function make_scorer from sklearn.
            Defaults to CUSTOM_SCORER_CLF.
        preprocessing (Union[Pipeline, str], optional): preprocessing
            pipeline to use. It can be None (no preprocessing), a
            predefined preprocessing pipeline ("default", ...) or
            a Pipeline object from sklearn. Defaults to "default":
            it applies a OneHotEncoder to "category" and object features,
            and it applies a SimpleImputer (median strategy) and a
            StandardScaler to numerical features.
        use_cross_validation (bool, optional): whether or not we apply
            a cross-validation. Defaults to True.
        cv (Any, optional): cross-validation object. It can be a predefined
            cross-validation setting ("default", "timeseries", ...), an
            iterable object, a cross-validation object from sklearn, etc.
            Defaults to "default": it applies a StratifiedShuffleSplit
            if a groups object is given when applying the fit method,
            otherwise, it uses a RepeatedKFold. In both cases, n_splits
            is set to 5.
        test_size (float, optional): test of the size set when splitting.
            Defaults to 0.2.
        shuffle (bool, optional): whether or not the algorithm shuffle the
            sample when splitting the given dataset. Defaults to False.
        sort_result_by (str, optional): on which metric do you want to
            sort the final dataframe. Defaults to "Balanced Accuracy".
        ascending (bool, optional): sort the final dataframe in
            ascending order?. Defaults to False.
        save_results (bool, optional): if True, save the results in a csv
            file. Defaults to False.
        path_results (str, optional): path to use when saving the results.
            Defaults to "".
        additional_model_params (Dict[str, Any], optional): dictionary
            that contains parameters to be applied to each element of the
            pipeline. E.g. {"n_estimators": 100}, apply to all the
            preprocessing tasks and/or models that have the parameter
            n_estimators with the parameter n_estimators. Defaults to {}.
        random_state (int, optional): random state variable. Is applied
            to every model and elements of the pipeline. Defaults to 42.
        verbose (int, optional): if greater than 1, print the warnings.
            Defaults to 0.
    Examples:

        >>> import pandas as pd
        >>> from sklearn.datasets import load_iris
        >>> from vulpes.automl import Classifiers

        >>> dataset = load_iris()
        >>> X = pd.DataFrame(dataset["data"], columns=dataset["feature_names"])
        >>> y = dataset["target"]

        >>> classifiers = Classifiers()
        >>> df_models = classifiers.fit(X, y)
        >>> df_models
        | Model                       | Balanced Accuracy | Accuracy | ...
        |-----------------------------|-------------------|----------|------
        Model									
        LinearDiscriminantAnalysis	    0.977625	        0.977333   ...
        MLPClassifier	                0.973753	        0.973333   ...
        QuadraticDiscriminantAnalysis	0.973219	        0.973333   ...
        KNeighborsClassifier	        0.971702	        0.969333   ...
        ...                             ...                 ...        ...
    """
    def __init__(
        self,
        *,
        models_to_try: Union[str, List[Tuple[str, Any]]] = "all",
        custom_scorer: Dict[str, Any] = CUSTOM_SCORER_CLF,
        preprocessing: Union[Pipeline, str] = "default",
        use_cross_validation: bool = True,
        cv: Any = "default",
        test_size: float = 0.2,
        shuffle: bool = False,
        sort_result_by: str = "Balanced Accuracy",
        ascending: bool = False,
        save_results: bool = False,
        path_results: str = "",
        additional_model_params: Dict[str, Any] = {},
        random_state: int = 42,
        verbose: int = 0,
    ):
        super().__init__()
        self.task = "classification"
        self.models_to_try = self.predefined_list_models(models_to_try)
        self.custom_scorer = custom_scorer
        self.preprocessing = preprocessing
        self.use_cross_validation = use_cross_validation
        self.cv = cv
        self.test_size = test_size
        self.shuffle = shuffle
        self.sort_result_by = sort_result_by
        self.ascending = ascending
        self.save_results = save_results
        self.path_results = path_results
        self.additional_model_params = additional_model_params
        self.random_state = random_state
        self.verbose = verbose

        if not (self.verbose):
            warnings.filterwarnings("ignore")

    def fit(
        self,
        X: Array_like,
        y: Array_like,
        *,
        sample_weight: Array_like = None,
        groups: Array_like = None,
    ) -> pd.DataFrame:
        """
        Fit many models

        Args:
            X (Array_like): Input dataset
            y (Array_like): Target variable
            sample_weight (Array_like, optional): sample weights
                to apply when fitting. Defaults to None.
            groups (Array_like, optional): groups to use during
                cross validation to stratify. Defaults to None.

        Raises:
            ValueError: cross validation wrong type, or failed
            RuntimeError: Error when fitting a model

        Returns:
            pd.DataFrame: dataframe with the goodness-of-fit metrics
                evaluated for each model.

        Examples:

        """
        # Convert X to dataframe
        # (some preprocessing task, model, etc require this format)
        if not (isinstance(X, pd.DataFrame)):
            X = pd.DataFrame(X)

        # dictionary to store calculated values, model info, etc for each model
        metrics_dic = defaultdict(list)

        # Fit and calculate metrics for all the models
        for name, model in tqdm(self.models_to_try):
            print("\n" + name)
            top = perf_counter()

            # extend the class with a custom predict proba function
            # based on a decision function or pairwise distances
            if not (hasattr(model, "predict_proba")):
                model = create_model_2(model)
            else:
                model = model()
            model_name = name.lower()

            # Create the pipeline
            pipe = self.create_pipeline(model_name, model)

            # When available:
            # - Activate "probability" parameter to calculate more metrics
            # - maximize n_jobs
            # - fix a random_state
            # - change the loss for SGDClassifier to allow the
            # predict_proba method
            model_params = {}
            for pipe_name, pipe_elt in pipe.steps:
                pipe_elt_available_params = pipe_elt.get_params().keys()
                if "random_state" in pipe_elt_available_params:
                    rd_state = f"{pipe_name}__random_state"
                    model_params[rd_state] = self.random_state
                if "normalize" in pipe_elt_available_params:
                    model_params[f"{pipe_name}__normalize"] = False
                if "n_jobs" in pipe_elt_available_params:
                    model_params[f"{pipe_name}__n_jobs"] = -1
                if "probability" in pipe_elt_available_params:
                    model_params[f"{pipe_name}__probability"] = True
                for (
                    add_param_name,
                    add_param_val,
                ) in self.additional_model_params.items():
                    if add_param_name in pipe_elt_available_params:
                        model_params[
                            f"{pipe_name}__{add_param_name}"
                        ] = add_param_val
            # change the loss to allow multiclass and predict proba
            if name == "SGDClassifier":
                model_params[f"{model_name}__loss"] = "log"
            if model_params != {}:
                pipe.set_params(**model_params)

            # if sample_weight is a parameter of the fit function, update
            # fit parameters
            fit_params = {}
            for pipe_name, pipe_elt in pipe.steps:
                if hasattr(pipe_elt, "fit") and (
                    "sample_weight" in pipe_elt.fit.__code__.co_varnames
                ):
                    fit_params[f"{pipe_name}__sample_weight"] = sample_weight

            if self.use_cross_validation:
                if not (
                    hasattr(self.cv, "split")
                    or isinstance(self.cv, numbers.Integral)
                    or isinstance(self.cv, Iterable)
                    or isinstance(self.cv, str)
                ):
                    raise ValueError(
                        "Expected cv as an integer, cross-validation "
                        "object (from sklearn.model_selection), "
                        "iterable or valid string predefined cv"
                    )
                if isinstance(self.cv, str):
                    cv = self.predefined_cv(X, groups)
                else:
                    cv = self.cv
                try:
                    cv_model = cross_validate(
                        pipe,
                        X,
                        y,
                        cv=cv,
                        return_estimator=True,
                        n_jobs=-1,
                        fit_params=fit_params,
                        scoring=self.custom_scorer,
                    )
                except ValueError as e:
                    print(e)
                    print("Cross validation failed. If groups provided, ")
                    print(
                        "maybe can't stratify because some groups "
                        "represented in the dataset don't contains enough"
                        " samples."
                    )
                    print("Using RepeatedKFold instead.")
                    cv = RepeatedKFold(
                        n_splits=5, n_repeats=5, random_state=self.random_state
                    )
                    cv_model = cross_validate(
                        pipe,
                        X,
                        y,
                        cv=cv,
                        return_estimator=True,
                        n_jobs=-1,
                        fit_params=fit_params,
                        scoring=self.custom_scorer,
                    )
                except Exception as e:
                    raise RuntimeError(str(e))

                # store fitted models
                self.fitted_models_[name] = cv_model["estimator"]

                # add metrics to the lists

                # nan mean are used, because when folding,
                # a label can be missing
                # in a particular fold, thus creating nan values
                for metric_name in self.custom_scorer.keys():
                    print_metric_name = METRIC_NAMES.get(metric_name,
                                                         metric_name)
                    (
                        metrics_dic[print_metric_name].append(
                            np.nanmean(cv_model[f"test_{metric_name}"])
                        )
                    )
                metrics_dic["Model"].append(name)  # add name
                # add running time
                metrics_dic["Running time"].append(perf_counter() - top)
            else:
                try:
                    X_train, X_test, y_train, y_test = train_test_split(
                        X,
                        y,
                        test_size=self.test_size,
                        shuffle=self.shuffle,
                        stratify=groups,
                        random_state=self.random_state,
                    )
                    pipe.fit(X_train, y_train, **fit_params)
                    res = _validation._score(
                        pipe, X_test, y_test, self.custom_scorer,
                        error_score="raise"
                    )
                except Exception as e:
                    raise RuntimeError(f"Error when fitting: {e}")

                # store fitted models
                self.fitted_models_[name] = pipe

                # add metrics to the lists

                # nan mean are used, because when folding,
                # a label can be missing
                # in a particular fold, thus creating nan values
                for metric_name in self.custom_scorer.keys():
                    print_metric_name = METRIC_NAMES.get(metric_name,
                                                         metric_name)
                    (
                        metrics_dic[print_metric_name].append(
                            np.nanmean(res[metric_name])
                        )
                    )
                metrics_dic["Model"].append(name)  # add name
                # add running time
                metrics_dic["Running time"].append(perf_counter() - top)

        # reverse these metrics (because lower is better)
        # ex: rmse, mae, mape
        for metric_name in METRICS_TO_REVERSE:
            if metric_name in metrics_dic:
                reverse_metric = [-x for x in metrics_dic[metric_name]]
                metrics_dic[metric_name] = reverse_metric
        # create a dataframe with the results
        df_models = pd.DataFrame.from_dict(metrics_dic)
        df_models = df_models.sort_values(
            by=self.sort_result_by, ascending=self.ascending
        ).set_index("Model")
        self.df_models = df_models

        if self.save_results:  # save results
            df_models.to_csv(
                opj.path.join(self.path_results,
                              f"vulpes_results_{self.task}.csv")
            )
        return df_models



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test_classifiers.py: Test the classifiers module
"""

from vulpes.automl import Classifiers

import numpy as np


def test_classifiers_1() -> None:
    # binary output + hard voting
    X = np.array(
        [
            [
                np.random.randint(-20, 20) * np.random.random(),
                np.random.randint(-10, 10) * np.random.random(),
            ]
            for _ in range(50)
        ]
    )
    y = np.array([np.random.randint(0, 2) for _ in range(50)])
    classifiers = Classifiers()
    _ = classifiers.fit(X, y)
    _ = classifiers.build_best_models(X, y, voting="hard")


def test_classifiers_2() -> None:
    # 3-class + soft voting + sample weight + groups
    X = np.array(
        [
            [
                np.random.randint(-20, 20) * np.random.random(),
                np.random.randint(-10, 10) * np.random.random(),
            ]
            for _ in range(50)
        ]
    )
    y = np.array([np.random.randint(0, 3) for _ in range(50)])
    sample_weight = np.array([np.random.random() for _ in range(50)])
    groups = np.array([np.random.randint(0, 2) for _ in range(50)])
    classifiers = Classifiers()
    _ = classifiers.fit(X, y, sample_weight=sample_weight, groups=groups)
    _ = classifiers.build_best_models(X, y, voting="hard")


def test_classifiers_3() -> None:
    # default preprocessing, no cross validation
    # + test get_ functions to retrieve models and predict functions
    # on fitted models
    X = np.array(
        [
            [
                np.random.randint(-20, 20) * np.random.random(),
                np.random.randint(-10, 10) * np.random.random(),
            ]
            for _ in range(50)
        ]
    )
    y = np.array([np.random.randint(0, 2) for _ in range(50)])
    classifiers = Classifiers(preprocessing="default", use_cross_validation=False)
    _ = classifiers.fit(X, y)
    _ = classifiers.build_best_models(X, y, voting="soft")

    _ = classifiers.get_fitted_models()
    _ = classifiers.get_best_model()
    _ = classifiers.predict(X)
    _ = classifiers.predict_proba(X)
    _ = classifiers.predict_best(X)
    _ = classifiers.predict_proba_best(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test_regressions.py: Test the regressions module
"""

from vulpes.automl import Regressions

import numpy as np


def test_regressions_1() -> None:
    # regression with "timeseries" predefined cross validation
    X = np.array(
        [
            [
                np.random.randint(-20, 20) * np.random.random(),
                np.random.randint(-10, 10) * np.random.random(),
            ]
            for _ in range(50)
        ]
    )
    y = np.array([np.random.randint(-5, 5) * np.random.random() for _ in range(50)])
    regressions = Regressions(cv="timeseries")
    _ = regressions.fit(X, y)
    _ = regressions.build_best_models(X, y)



import vulpes

from setuptools import setup, find_packages

VERSION = vulpes.__version__
DESCRIPTION = ("Test many classification, regression models and "
               "clustering algorithms to see which one is most "
               "suitable for your dataset.")
# LONG_DESCRIPTION: README + CHANGELOG
LONG_DESCRIPTION = ""
with open("README.md", "r+", encoding="UTF-8") as f:
    LONG_DESCRIPTION += f.read()
LONG_DESCRIPTION += "\n\n"
with open("CHANGELOG.rst", "r+") as f:
    LONG_DESCRIPTION += f.read()
# Install Requires
with open("requirements.txt", "r+") as f:
    INSTALL_REQUIRES = [x.replace("\n", "") for x in f.readlines()]

URL = "https://vulpes.readthedocs.io/en/latest/"
DOWNLOAD_URL = "https://pypi.org/project/vulpes/"
PROJECT_URLS = {
    "Bug Tracker": "https://github.com/AdrienC21/vulpes/issues",
    "Documentation": URL,
    "Source Code": "https://github.com/AdrienC21/vulpes",
}
AUTHOR = "Adrien Carrel"
AUTHOR_EMAIL = "a.carrel@hotmail.fr"
PYTHON_REQUIRES = ">=3.7"
CLASSIFIERS = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Unix',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules'
]
KEYWORDS = ["vulpes", "python", "automl", "scikit-learn",
            "machine-learning", "machine", "learning", "model",
            "artificial", "intelligence", "clustering",
            "dataset", "classification", "regression",
            "hyperparameter-tuning"]

setup(
    name="vulpes",
    version=VERSION,
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    url=URL,
    download_url=DOWNLOAD_URL,
    project_urls=PROJECT_URLS,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR,
    maintainer_email=AUTHOR_EMAIL,
    license="MIT",
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=find_packages(),
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""corevulpes.py: Parent class that contains
common methods shared between children classes
"""

from ..utils.utils import (
    CLASSIFIERS,
    REGRESSIONS,
    CLUSTERING,
    METRIC_NAMES,
    METRICS_TO_REVERSE,
    create_model_2,
)

import warnings
import numbers
import operator
from functools import reduce
from time import perf_counter
from typing import List, Dict, Any, Union, Tuple
from collections import defaultdict
from collections.abc import Iterable
from abc import ABC

import numpy as np
import pandas as pd
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.pipeline import Pipeline
from sklearn.model_selection import (
    cross_validate,
    RepeatedKFold,
    StratifiedShuffleSplit,
    _validation,
    train_test_split,
    TimeSeriesSplit,
)
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.compose import ColumnTransformer, make_column_selector as selector
from sklearn.impute import SimpleImputer
from sklearn.exceptions import NotFittedError
from sklearn.metrics._scorer import _ProbaScorer, _PredictScorer

warnings.filterwarnings("ignore")
# define type Array_like
Array_like = Union[List, pd.DataFrame, pd.Series, np.ndarray, Any]


class CoreVulpes(ABC):
    """
    Parent class with shared methods between the classes
    Classifiers, Regressions and Clustering
    """

    def __init__(self):
        # store the metrics dataframe after fitting
        self.df_models = None  # all the models
        self.df_best_model = None  # the best one with a voting clf/reg
        # store the fitted models
        self.fitted_models_ = {}  # all the models
        self.best_model_ = None  # the best one with a voting clf/reg

    def missing_data(self, X: Array_like) -> pd.DataFrame:
        """
        Evaluate the absolute count and the percentage of
        missing data in a particular dataset

        Args:
            X (Array_like): Dataset

        Returns:
            pd.DataFrame: Absolute count and percentage of missing
            data in X

        Examples:
            >>> import pandas as pd
            >>> import numpy as np
            >>> df = pd.DataFrame([["a", "x"],
                                   [np.nan, "y"],
                                   ["a", np.nan],
                                   ["b", np.nan]],
                                  dtype="category",
                                  columns=["feature1", "feature2"])
            >>> classifiers.missing_data(df)
            | Total Missing | Percentage (%) | Accuracy |
            |--------------:|---------------:|---------:|
            |    feature2   |              2 |     50.0 |
            |    feature1   |              1 |     25.0 |
        """
        if not (isinstance(X, pd.DataFrame)):
            X = pd.DataFrame(X)  # locally modify X
        total_missing = X.isnull().sum().sort_values(ascending=False)
        percent_missing = (X.isnull().sum() / X.isnull().count()).sort_values(
            ascending=False
        ) * 100
        missing = pd.concat(
            [total_missing, percent_missing],
            axis=1,
            keys=["Total Missing", "Percentage (%)"],
        )
        return missing

    def predefined_preprocessing(self) -> Pipeline:
        """
        Either return a predefined preprocessing pipeline
        if self.preprocessing is a string. Otherwise, check
        if self.preprocessing is in fact a Pipeline or a None
        object (in that case, no preprocessing applied).

        Raises:
            ValueError: Unknown string
            TypeError: self.preprocessing is not a string,
            not None, not a pipeline object

        Returns:
            Pipeline: None of preprocessing Pipeline to apply to each models

        Examples:

        """
        if isinstance(self.preprocessing, str):
            if self.preprocessing == "default":
                # Imputer + standard scaler for not categorical values
                numeric_transformer = Pipeline(
                    steps=[
                        ("imputer", SimpleImputer(strategy="median")),
                        ("scaler", StandardScaler()),
                    ]
                )
                # OneHotEncoder for categorical values
                categorical_transformer = OneHotEncoder(handle_unknown="ignore")

                preprocessing = ColumnTransformer(
                    transformers=[
                        (
                            "num",
                            numeric_transformer,
                            selector(dtype_exclude=["category", object]),
                        ),
                        (
                            "cat",
                            categorical_transformer,
                            selector(dtype_include=["category", object]),
                        ),
                    ]
                )

                preprocessing_pipeline = Pipeline(
                    steps=[("preprocessing", preprocessing)]
                )
            # Insert predefined preprocessing pipelines here!
            else:
                raise ValueError(
                    "Unknown parameter: preprocessing."
                    "Please enter a valid preprocessing "
                    "(an already implemented one, "
                    ", a Pipeline object, or None)."
                )
        # if it's not a string, not a pipeline, or not None, raise error
        elif not (
            isinstance(self.preprocessing, Pipeline) or (self.preprocessing is None)
        ):
            raise TypeError(
                "Preprocessing must be a string, " "a Pipeline object, or None."
            )
        else:  # return the given preprocessing pipeline or None object
            preprocessing_pipeline = self.preprocessing
        return preprocessing_pipeline

    def create_pipeline(self, model_name: str, model: Any) -> Pipeline:
        """
        Create a pipeline by combining an optional preprocessing
        pipeline and the given model

        Args:
            model_name (str): name of the model (lowercase)
            model (Any): Model at the end of the pipeline

        Returns:
            Pipeline: Pipeline with a preprocessing task (if not set to None)
            and the given model

        Examples:

        """
        # Preprocessing
        preprocessing_pipeline = self.predefined_preprocessing()

        # Model
        model_pipeline = Pipeline(steps=[(model_name, model)])

        # Merge all the steps
        pipelines = [preprocessing_pipeline, model_pipeline]
        steps = reduce(operator.add, [p.steps for p in pipelines if not (p is None)])
        return Pipeline(steps=steps)

    def predefined_cv(self, X: Array_like = None, groups: Array_like = None) -> Any:
        """
        Convert a cross validation string (self.cv parameter)
        into a predefined cross validation object

        Args:
            X (Array_like, optional): if necessary, X is the dataset.
            Defaults to None.
            groups (Array_like, optional): if necessary, groups is
            an array-like object on which we stratify to create
            the different folds. Defaults to None.

        Raises:
            ValueError: raise an error if the string doesn't correspond
            to a predefined cross validation

        Returns:
            Any: Cross validation object

        Examples:

        """
        if self.cv == "default":
            # if groups, cross validation is a
            # stratified shuffle, else,
            # a repeatedKFold
            if groups is None:
                cv = RepeatedKFold(
                    n_splits=5, n_repeats=5, random_state=self.random_state
                )
            else:
                sss = StratifiedShuffleSplit(
                    n_splits=5, test_size=self.test_size, random_state=self.random_state
                )
                cv = sss.split(X, groups)
        elif self.cv == "timeseries":
            cv = TimeSeriesSplit(n_splits=5)

        # Insert predefined cross validation here!
        else:
            raise ValueError(f"Unknown cross validation: {self.cv}")
        return cv

    def predefined_list_models(
        self, models_to_try: Union[str, List[Tuple[str, Any]]] = "all"
    ) -> List[Tuple[str, Any]]:
        """
        If models_to_try isn't a list of models but is a string,
        return the corresponding predefined list of models to test

        Args:
            models_to_try (Union[str, List[Tuple[str, Any]]],
                           optional): string of predefined list of models
            or list of tuple (name of model, model). Defaults to "all".

        Raises:
            ValueError: raise an error if models_to_try is a string that
            doesn't correspond to any predefined list of models

        Returns:
            List[Tuple[str, Any]]: list of tuple (name of the model, model)

        Examples:

        """
        if not (isinstance(models_to_try, str)):
            return models_to_try
        # else, it's a string, search for predefined list of models
        if models_to_try == "all":
            if self.task == "classification":
                return CLASSIFIERS
            elif self.task == "regression":
                return REGRESSIONS
            elif self.task == "clustering":
                return CLUSTERING
        # Insert new predefined list here!
        else:
            raise ValueError(
                f"Unknown parameter models_to_try: {models_to_try}. "
                "Please enter a valid list of models (tuple like "
                '("XGBClassifier", xgboost.XGBClassifier)) or an '
                'existing predefined list of models("all", ...)'
            )

    def remove_proba_metrics(
        self, dic_scorer: Dict[str, Union[_ProbaScorer, _PredictScorer]]
    ) -> Dict[str, Union[_ProbaScorer, _PredictScorer]]:
        """
        Take a dictionnary of metrics to evaluate the goodness-of-fit of
        classifiers as an input. Return a new version of this
        dictionnary with only the metrics that don't need
        probabilities to be calculated (e.g. AUROC)

        Args:
            dic_scorer (Dict[str, Union[_ProbaScorer, _PredictScorer]]):
            dictionnary of metrics

        Returns:
            Dict[str, Union[_ProbaScorer, _PredictScorer]]: filtered
            dictionnary of metrics

        Examples:

        """
        new_dic = {}
        for scorer_name, scorer in dic_scorer.items():
            if isinstance(scorer, _PredictScorer):  # no predict proba
                new_dic[scorer_name] = scorer
        return new_dic

    def build_best_models(
        self,
        X: Array_like,
        y: Array_like,
        *,
        sample_weight: Array_like = None,
        groups: Array_like = None,
        nb_models: int = 5,
        sort_result_by: str = None,
        ascending: bool = None,
        voting: str = "hard",
        weights: Array_like = None,
    ) -> pd.DataFrame:
        """
        When many models have been fitted, create an aggregated model
        using a voting system by selecting the nb_models best models based on
        the metric sort_result_by.

        Args:
            X (Array_like): dataset to fit the 'best model'
            y (Array_like): response/outcome variable
            sample_weight (Array_like, optional): sample weight.
                Defaults to None.
            groups (Array_like, optional): groups to stratify during
                cross validation. Defaults to None.
            nb_models (int, optional): number of models to select when
                creating the aggregated model. Defaults to 5.
            sort_result_by (str, optional): metrics to evaluate the best
                models that will be selected among the ones that we trained.
                Defaults to None.
            ascending (bool, optional): if ascending=True, the lower the
                metric sort_result_by is, the better the model is.
                Defaults to None.
            voting (str, optional): "hard" or "soft". If "soft", use
                the predicted probabilities of the different estimators
                to make a prediction. Defaults to "hard".
            weights (Array_like, optional): attribute different weights
                to each estimators. Defaults to None, which is equal to
                equal weights.

        Raises:
            ValueError: Voting Clustering not available
            NotFittedError: Fit models before building a 'best model'
            ValueError: less fitted models that the parameter nb_models
            NotImplementedError: Voting Clustering not available
            NotImplementedError: Sample weight not available
            ValueError: Wrong type of cross validation
            RuntimeError: Error when fitting an estimator

        Returns:
            pd.DataFrame: Performance of the aggregated model
                on different metrics

        Examples:

        """
        if self.task == "clustering":
            raise ValueError("Can't create a Voting Clustering algorithm.")
        if self.df_models is None:
            raise NotFittedError(
                "Please fit the models first by calling the method .fit "
                "before building an ensemble model"
            )
        if nb_models > len(self.df_models):
            raise ValueError(
                f"Not enough trained models ({len(self.df_models)}) "
                f"to select the best {nb_models} ones"
            )

        top = perf_counter()  # start to measure fitting time

        # Convert X to dataframe
        # (some preprocessing task, model, etc require this format)
        if not (isinstance(X, pd.DataFrame)):
            X = pd.DataFrame(X)

        # if undefined, take default values for sort_result_by and ascending
        # to select the best models
        if sort_result_by is None:
            sort_result_by = self.sort_result_by
        if ascending is None:
            ascending = self.ascending
        # sort the models based on the given metric
        sorted_df_models = self.df_models.sort_values(
            by=sort_result_by, ascending=ascending
        )
        # name of the best models
        best_models_names = list(sorted_df_models.index)[:nb_models]
        # dictionnary: model name -> corresponding class
        dict_models = dict(self.predefined_list_models(self.models_to_try))
        # list of tuple (model name, instance of model)

        # create the list of estimators
        if self.task == "classification":  # check predict_proba method
            estimators = []
            for b in best_models_names:
                if not (hasattr(dict_models[b], "predict_proba")):
                    estimator = create_model_2(dict_models[b])
                else:
                    estimator = dict_models[b]()
                estimators.append((b.lower(), estimator))
        else:
            estimators = [(b.lower(), dict_models[b]()) for b in best_models_names]

        # define the voting model
        if self.task == "classification":
            voting = VotingClassifier(
                estimators, voting=voting, weights=weights, n_jobs=-1
            )
        elif self.task == "regression":
            voting = VotingRegressor(estimators, weights=weights, n_jobs=-1)
        else:
            raise NotImplementedError("Unknown task: {self.task}")
        voting_name = voting.__class__.__name__.lower()
        # Add preprocessing, create a pipeline
        pipe = self.create_pipeline(voting_name, voting)
        # adjust some hyperparameters when available for each model inside the
        # voting model and the pipeline
        model_params = {}
        for pipe_name, pipe_elt in pipe.steps:
            pipe_elt_available_params = pipe_elt.get_params().keys()
            if "random_state" in pipe_elt_available_params:
                model_params[f"{pipe_name}" "__random_state"] = self.random_state
            if "normalize" in pipe_elt_available_params:
                model_params[f"{pipe_name}__normalize"] = False
            if "n_jobs" in pipe_elt_available_params:
                model_params[f"{pipe_name}__n_jobs"] = -1
            for add_param_name, add_param_val in self.additional_model_params.items():
                if add_param_name in pipe_elt_available_params:
                    model_params[f"{pipe_name}" f"__{add_param_name}"] = add_param_val

        for (model_name, model_instance) in estimators:
            model_available_params = model_instance.get_params().keys()
            if "random_state" in model_available_params:
                model_params[
                    f"{voting_name}__{model_name}__" "random_state"
                ] = self.random_state
            if "n_jobs" in model_available_params:
                model_params[f"{voting_name}__{model_name}__n_jobs"] = -1
            if "probability" in model_available_params:
                model_params[f"{voting_name}__{model_name}__" "probability"] = True
            if "normalize" in model_available_params:
                model_params[f"{voting_name}__{model_name}__" "normalize"] = False
            for add_param_name, add_param_val in self.additional_model_params.items():
                if add_param_name in model_available_params:
                    model_params[
                        f"{voting_name}__{pipe_name}" f"__{add_param_name}"
                    ] = add_param_val
            # change the loss to allow multiclass and predict proba
            if model_name == "SGDClassifier".lower():
                model_params[f"{voting_name}__{model_name}__loss"] = "log"
        if model_params != {}:
            pipe.set_params(**model_params)

        ######
        # Voting Classifier / Regressor don't really support sample weight
        # all of the models need a sample weight :/
        if not (sample_weight is None):
            raise NotImplementedError(
                "Sample weight not (yet) implemented when fitting a" " 'best model'."
            )
        fit_params = {}
        ######

        # dictionary to store calculated values, model info, etc for each model
        metrics_dic = defaultdict(list)

        # if it's a voting classifier with "hard" voting, as it
        # doesn't (yet) allow the predict proba method, we need to
        # remove some metrics when evaluating the performance
        # of the model
        if (self.task == "classification") and (
            pipe[-1].get_params()["voting"] == "hard"
        ):
            scoring = self.remove_proba_metrics(self.custom_scorer)
        else:
            scoring = self.custom_scorer

        if self.use_cross_validation:
            if not (
                hasattr(self.cv, "split")
                or isinstance(self.cv, numbers.Integral)
                or isinstance(self.cv, Iterable)
                or isinstance(self.cv, str)
            ):
                raise ValueError(
                    "Expected cv as an integer, cross-validation "
                    "object (from sklearn.model_selection), "
                    "iterable or valid string predefined cv"
                )
            if isinstance(self.cv, str):
                cv = self.predefined_cv(X, groups)
            else:
                cv = self.cv
            try:
                cv_model = cross_validate(
                    pipe,
                    X,
                    y,
                    cv=cv,
                    return_estimator=True,
                    n_jobs=-1,
                    fit_params=fit_params,
                    scoring=scoring,
                )
            except ValueError as e:
                print(e)
                print("Cross validation failed. If groups provided, ")
                print(
                    "maybe can't stratify because some groups "
                    "represented in the dataset don't contains enough"
                    " samples."
                )
                print("Using RepeatedKFold instead.")
                cv = RepeatedKFold(
                    n_splits=5, n_repeats=5, random_state=self.random_state
                )
                cv_model = cross_validate(
                    pipe,
                    X,
                    y,
                    cv=cv,
                    return_estimator=True,
                    n_jobs=-1,
                    fit_params=fit_params,
                    scoring=scoring,
                )
            except Exception as e:
                raise RuntimeError(str(e))

            # add metrics to the lists

            # nan mean are used, because when folding,
            # a label can be missing
            # in a particular fold, thus creating nan values
            for metric_name in scoring.keys():
                print_metric_name = METRIC_NAMES.get(metric_name, metric_name)
                (
                    metrics_dic[print_metric_name].append(
                        np.nanmean(cv_model[f"test_{metric_name}"])
                    )
                )

            metrics_dic["Model"].append(f"Voting ({nb_models}-best)")
            # add running time
            metrics_dic["Running time"].append(perf_counter() - top)
        else:
            try:
                X_train, X_test, y_train, y_test = train_test_split(
                    X,
                    y,
                    test_size=self.test_size,
                    shuffle=self.shuffle,
                    stratify=groups,
                    random_state=self.random_state,
                )
                pipe.fit(X_train, y_train, **fit_params)
                res = _validation._score(
                    pipe, X_test, y_test, scoring, error_score="raise"
                )
            except Exception as e:
                raise RuntimeError(f"Error when fitting: {e}")

            # add metrics to the lists

            # nan mean are used, because when folding,
            # a label can be missing
            # in a particular fold, thus creating nan values
            for metric_name in scoring.keys():
                print_metric_name = METRIC_NAMES.get(metric_name, metric_name)
                (metrics_dic[print_metric_name].append(np.nanmean(res[metric_name])))
            metrics_dic["Model"].append(f"Voting ({nb_models}-best)")
            # add running time
            metrics_dic["Running time"].append(perf_counter() - top)

        # reverse these metrics (because lower is better)
        # ex: rmse, mae, mape
        for metric_name in METRICS_TO_REVERSE:
            if metric_name in metrics_dic:
                metrics_dic[metric_name] = [-x for x in metrics_dic[metric_name]]

        # Metric for the fitted "best model"
        df_best_model = pd.DataFrame.from_dict(metrics_dic)
        df_best_model = df_best_model.set_index("Model")
        self.best_model_ = pipe
        self.df_best_model = df_best_model

        return df_best_model

    def get_fitted_models(self) -> Dict[str, Union[Pipeline, List[Pipeline]]]:
        """
        Get a dictionnary with the fitted models

        Raises:
            NotFittedError: Models have not been fitted yet

        Returns:
            Dict[str, Union[Pipeline, List[Pipeline]]]: Dictionnary
            with, for all models, either the fitted model, or all
            the fitted models (during cross validation).

        Examples:

        """

        if self.fitted_models_ is None:
            raise NotFittedError(
                "Fit some models before retrieving them by calling " "the method .fit"
            )
        return self.fitted_models_

    def get_best_model(self) -> Pipeline:
        """
        Get the model created and fitted by the
        method build_best_models

        Raises:
            NotFittedError: Models not trained
            NotFittedError: Best model not calculated

        Returns:
            Pipeline: 'Best model' using multiple fitted models

        Examples:

        """
        if self.best_model_ is None:
            if self.fitted_models_ is None:
                raise NotFittedError(
                    "Fit some models before retrieving them by calling "
                    "the method .fit"
                )
            raise NotFittedError(
                "Many models have been fitted. But the 'best model' "
                "hasn't been fitted yet. Please call the method "
                ".build_best_models before retrieving it."
            )
        return self.best_model_

    def predict(
        self, X: Array_like, *, dataframe_format: bool = True
    ) -> Union[pd.DataFrame, Dict[str, np.ndarray]]:
        """
        Evaluate all the fitted models on the dataset X

        Args:
            X (Array_like): Array-like object on which we'll make prediction(s)
            dataframe_format (bool, optional): if True, then the result
            is a dataframe with all the predictions for all the models.
            Defaults to True.

        Returns:
            Union[pd.DataFrame, Dict[str, np.ndarray]]: Dictionnary or
            dataframe with the predictions

        Examples:

        """
        fitted_models = self.get_fitted_models()
        res = {}
        for name, fitted_model in fitted_models.items():
            if isinstance(fitted_model, List) or isinstance(fitted_model, np.ndarray):
                print(
                    "Cross validation used to fit. Select the model"
                    " fitted on the first fold."
                )
                fitted_model = fitted_model[0]
            res[name] = fitted_model.predict(X)
        if not (dataframe_format):
            return res
        return pd.DataFrame.from_dict(res)

    def predict_proba(
        self,
        X: Array_like,
    ) -> Dict[str, np.ndarray]:
        """
        Based on the fitted models, make many probability
        predictions on the dataset X

        Args:
            X (Array_like): Dataset

        Returns:
            Dict[str, np.ndarray]: dictionnary with, for each model,
            an array of the corresponding predicted probabilities

        Examples:

        """
        if self.task == "clustering":
            return NotImplementedError(
                "predict_proba method not (yet) implemented for "
                "clustering algorithms"
            )
        fitted_models = self.get_fitted_models()
        res = {}
        for name, fitted_model in fitted_models.items():
            if isinstance(fitted_model, List) or isinstance(fitted_model, np.ndarray):
                print(
                    "Cross validation used to fit. Select the model"
                    " fitted on the first fold."
                )
                fitted_model = fitted_model[0]
            res[name] = fitted_model.predict_proba(X)
        return res

    def predict_best(self, X: Array_like) -> np.ndarray:
        """
        Evaluate the fitted 'best model' on the array-like X

        Args:
            X (Array_like): dataset

        Returns:
            np.ndarray: array of predictions

        Examples:

        """
        best_model = self.get_best_model()
        if isinstance(best_model, List) or isinstance(best_model, np.ndarray):
            print(
                "Cross validation used to fit. Select the model"
                " fitted on the first fold."
            )
            best_model = best_model[0]
        return best_model.predict(X)

    def predict_proba_best(self, X: Array_like) -> np.ndarray:
        """
        Evaluate the fitted 'best model' on the array-like X
        and return probabilities

        Args:
            X (Array_like): dataset

        Returns:
            np.ndarray: predicted probabilities

        Examples:

        """
        best_model = self.get_best_model()
        if isinstance(best_model, List) or isinstance(best_model, np.ndarray):
            print(
                "Cross validation used to fit. Select the model"
                " fitted on the first fold."
            )
            best_model = best_model[0]
        return best_model.predict_proba(X)
